﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Net.Http;
using System.Security.Cryptography;
using System.ServiceModel;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Xml.Linq;
using UserStageExample.FlexiCapture;

namespace UserStageExample.FlexiService
{
    /// <summary>
    /// A connector to FlexiCapture
    /// </summary>
    internal class FlexiConnectorService : IDisposable
    {
        /// <summary>
        /// The path to the FlexiCapture API
        /// </summary>
        private const string FlexiHandlerPath = "/FlexiCapture12/Server/{0}API/v1/Soap";
        /// <summary>
        /// The path to authenticate FlexiCapture
        /// </summary>
        private const string FlexiCaptureAuthPart = "FCAuth/";
        /// <summary>
        /// The allowable size of a message sent to API
        /// </summary>
        private const int MessageSize = 4*1024*1024;
        /// <summary>
        /// The name of the project used in the example
        /// </summary>
        private const string UserStageProjectName = "UserStageExample";
        /// <summary>
        /// The name of the service stage
        /// </summary>
        private const string UserStageName = "Service";
        
        /// <summary>
        /// The user role is the User station Operator
        /// </summary>
        private const int RoleType = 12;
        /// <summary>
        /// The station type is the User Station
        /// </summary>
        private const int WorkstationType = 10;
        /// <summary>
        /// The type of the Exceptions stage
        /// </summary>
        private const int ExceptionStageType = 1000;
        /// <summary>
        /// The file that contains recognized data
        /// </summary>
        private const string DocumentBodyFileName = "DocumentBody";

        /// <summary>
        /// The name of the file with the electronic signature which is added to confirmed documents
        /// </summary>
        private const string SignatureFileName = "Signature.dat";
        /// <summary>
        /// The size of the electronic signature which is added to documents
        /// </summary>
        private const int SignatureSize = 1024;
        /// <summary>
        /// The type of the attachment is the document attachment
        /// </summary>
        private const int DocumentAttachment = 9;

        /// <summary>
        /// The reason for rejection which is added to documents
        /// </summary>
        private const string RejectReason = "Rejecting, moving to exceptions";
        /// <summary>
        /// At the processing stage the completion of the current stage is confirmed.
        /// The batch is sent further in accordance with the FlexiCapture scenario.
        /// </summary>
        private const int CommitUserStageType = 410;

        private const int CloseUserStageType = -1;

        /// <summary>
        /// A SOAP client
        /// </summary>
        private readonly FlexiCaptureWebServiceSoapClient _client;

        /// <summary>
        /// The ID of a session opened in FlexiCapture
        /// </summary>
        private readonly int _sessionId = 0;

        /// <summary>
        /// The ID of the FlexiCapture project
        /// </summary>
        private readonly int _projectId = 0;

        /// <summary>
        /// The ID of the FlexiCapture service stage
        /// </summary>
        private readonly int _serviceStageId = 0;

        /// <summary>
        /// The ID of the FlexiCapture exceptions stage
        /// </summary>
        private readonly int _exceptionsStageId = 0;

        /// <summary>
        /// A sign of connector's disposition
        /// </summary>
        private int _disposed = 0;

        /// <summary>
        /// The generator of digital signatures
        /// </summary>
        private readonly RandomNumberGenerator _rng = new RNGCryptoServiceProvider();


        /// <summary>
        /// Creation of a connector
        /// </summary>
        /// <param name="serverUri">The path to the FlexiCapture server</param>
        /// <param name="windowsAuth">Whether the Negotiate authentication should be used</param>
        /// <param name="login">A login is ignored if windowsAuth == true</param>
        /// <param name="password">A password is ignored if windowsAuth == true</param>
        public FlexiConnectorService(Uri serverUri, bool windowsAuth, string login, string password)
        {
            // Creating a new binding
            var binding = new BasicHttpBinding();

            // The size of a default message is too small for FlexiCapture and must me increased
            binding.MaxReceivedMessageSize = MessageSize;

            // Selecting a connection security mode depending on a protocol 
            binding.Security.Mode = serverUri.Scheme == Uri.UriSchemeHttps
                ? BasicHttpSecurityMode.Transport
                : BasicHttpSecurityMode.TransportCredentialOnly;

            // Selecting an authentiction method
            binding.Security.Transport.ClientCredentialType = windowsAuth
                ? HttpClientCredentialType.Windows
                : HttpClientCredentialType.Basic;

            // Selecting a path to API depending on the authentication method
            var externalSuffix = windowsAuth ? string.Empty : FlexiCaptureAuthPart;
            var remoteAddress = new EndpointAddress(new Uri(serverUri, string.Format(FlexiHandlerPath, externalSuffix)));

            // Creating a SOAP client
            _client = new FlexiCapture.FlexiCaptureWebServiceSoapClient(binding, remoteAddress);
            // Specifying a login and password to authenticate FlexiCapture
            if (_client.ClientCredentials != null && !windowsAuth)
            {
                _client.ClientCredentials.UserName.UserName = login;
                _client.ClientCredentials.UserName.Password = password;
            }

            try
            {
                // Opening a session
                _sessionId = _client.OpenSession(RoleType, WorkstationType);
                var projectGuid = string.Empty;
                // Looking for the required project
                foreach (var project in _client.GetProjects())
                {
                    if (project.Name == UserStageProjectName)
                    {
                        projectGuid = project.Guid;
                    }
                }
                if (string.IsNullOrEmpty(projectGuid))
                {
                    MessageBox.Show("The server does not contain the UserStageExample project required by the example");
                }
                // Opening the project
                _projectId = _client.OpenProject(_sessionId, projectGuid);

                // Finding and saving the service stage ID
                var stages = _client.GetProcessingStages(_projectId, 0, 0, UserStageName);
                if (stages.Count != 1) throw new Exception("Invalid project");
                _serviceStageId = stages[0].Id;

                // Finding and saving the External ID of the exceptions stage
                stages = _client.GetProcessingStages(_projectId, 0, ExceptionStageType, string.Empty);
                if (stages.Count != 1) throw new Exception("Invalid project");
                _exceptionsStageId = stages[0].ExternalId;
            }
            catch (Exception)
            {
                Dispose();
                throw;
            }
        }

        /// <summary>
        /// Returning the list of tasks that are pending processing at the service stage
        /// </summary>
        /// <returns></returns>
        public Dictionary<int, Task> GetTasks()
        {
            // Getting the list of tasks
            var tasks = _client.GetAvailableTasksByStageId(_sessionId, _projectId, 
                new FlexiCapture.stageIds() { _serviceStageId }, false);

            var result = new Dictionary<int, Task>();
            if (tasks == null || tasks.Count == 0) return result;

            foreach (var task in tasks)
            {
                // Creating an object of the Task type for each task
                var item = new Task
                {
                    Id = task.Id,
                    BatchId = task.BatchId
                };

                // Getting the list of documents included in the task
                var documents = _client.GetTaskDocuments(task.Id);
                if (documents == null || documents.Count == 0) continue;
                foreach (var document in documents)
                {
                    // Loading the supporting XML
                    var file = _client.LoadDocumentResult(_sessionId, document.BatchId, document.Id, DocumentBodyFileName);
                    if (file == null || file.Bytes == null) continue;
                    // Disassembling XML
                    var doc = new Document(document.Id, XDocument.Load(new MemoryStream(file.Bytes)));
                    // Saving the disassembled list of documents
                    item.Documents.Add(document.Id, doc);
                }
                result.Add(item.Id, item);
            }
            return result;
        }

        /// <summary>
        /// Launching the task for further processing
        /// </summary>
        /// <param name="batchId">A batch ID</param>
        /// <param name="taskId">A task ID</param>
        /// <param name="acceptedDocuments">A list of documents to be confirmed</param>
        /// <param name="rejectedDocuments">A list of documents that are sent to the exceptions stage</param>
        public void Process(int batchId, int taskId, IEnumerable<Document> acceptedDocuments, IEnumerable<Document> rejectedDocuments)
        {
            if (!_client.OpenTask(_sessionId, taskId))
            {
                MessageBox.Show("Task is locked");
                return;
            }
            var ids = new FlexiCapture.docIds();
            foreach (var rejectedDocument in rejectedDocuments)
            {
                ids.Add(rejectedDocument.Id);
            }
            if (ids.Count > 0)
            {
                _client.CreateTask(_sessionId, batchId, _exceptionsStageId, 0, RejectReason, 0, ids, false);
            }
            _client.OpenBatch(_sessionId, batchId);
            var hasDocuments = false;
            try
            {
                var uri = _client.Endpoint.Address.Uri.OriginalString.Replace("/API/v1/Soap", "/FileService/v1");
                foreach (var document in acceptedDocuments)
                {
                    hasDocuments = true;

                    var buffer = new byte[SignatureSize];
                    _rng.GetBytes(buffer);
                    _client.SaveAttachment(_sessionId, DocumentAttachment, document.Id, batchId, _projectId, new FlexiCapture.File
                    {
                        Bytes = buffer,
                        Name = SignatureFileName
                    });
                    document.CreateStamp();
                    UploadTextFileToServer(new Uri(uri),
                        _client.ClientCredentials.UserName.UserName,
                        _client.ClientCredentials.UserName.Password,
                        document.GetXml(), 0, _sessionId, _projectId,
                        batchId, 0, document.Id, 0, "DocumentBody"
                    ).Wait();
                }
            }
            finally
            {
                _client.CloseBatch(_sessionId, batchId);
            }
            _client.CloseTask(_sessionId, taskId, hasDocuments ? CommitUserStageType : CloseUserStageType);
        }


        private static async Task<string> UploadTextFileToServer(Uri server, string login, string password,
            string stream, int objectType, int sessionId, int projectId, int batchId, int parentId,
            int objectId, int version, string streamName)
        {
            var creds = CredentialCache.DefaultNetworkCredentials;
            if (!string.IsNullOrEmpty(login)) creds = new NetworkCredential(login, password);
            using (var handler = new HttpClientHandler { Credentials = creds })
            using (var client = new HttpClient(handler))
            {
                var encodedStreamName = Convert.ToBase64String(Encoding.Unicode.GetBytes(streamName));
                var content = new MultipartFormDataContent
                {
                    {new StringContent("Save"), "Action"},
                    {new StringContent(objectType.ToString()), "objectType"},
                    {new StringContent(sessionId.ToString()), "sessionId"},
                    {new StringContent(projectId.ToString()), "projectId"},
                    {new StringContent(batchId.ToString()), "batchId"},
                    {new StringContent(parentId.ToString()), "parentId"},
                    {new StringContent(objectId.ToString()), "objectId"},
                    {new StringContent(version.ToString()), "version"},
                    {new StringContent(encodedStreamName), "streamName"},
                    {new ByteArrayContent(Encoding.Unicode.GetBytes(stream)), "blob", "data.txt"}
                };
                var msg = client.PostAsync(server, content).Result;
                return await msg.Content.ReadAsStringAsync();
            }
        }


        /// <summary>
        /// Disposition of the connector
        /// </summary>
        public void Dispose()
        {
            // Thanks to Interlocked only one flow will start cleaning up
            if (Interlocked.Exchange(ref _disposed, 1) != 0) return;
            // The first stage guarantees only one clean up flow. No need to syncrhonize once again
            if (_projectId != 0)
            {
                _client.CloseProject(_sessionId, _projectId);
            }
            if (_sessionId != 0)
            {
                _client.CloseSession(_sessionId);
            }
        }
    }
}
